import threading
from queue import Queue
from typing import Any


class Pipe:
    """A threadsafe passage to send data between threads."""
    def __init__(self):
        self.__queue = Queue(1)
        self.__lock = threading.Lock()
        self.__recv_check = threading.Condition(self.__lock)
        self.__send_check = threading.Condition(self.__lock)

    def send(self, value: Any):
        """Send data through the pipe."""
        with self.__send_check:
            while self.__queue.full():
                self.__send_check.wait()

            self.__queue.put(value)
            self.__recv_check.notify()

    def recv(self):
        """Receive data from the pipe."""
        with self.__recv_check:
            while self.__queue.empty():
                self.__recv_check.wait()

            value = self.__queue.get()
            self.__send_check.notify()

            return value



import time
from threading import Thread, Lock
from typing import Callable, Union

from BetterThreads.PooledThread import PooledThread


class ThreadPool:
    def __init__(self):
        """A simple thread manager."""
        self.__lock: Lock = Lock()
        self.__threads: dict[Callable, PooledThread] = dict()
        self.__dead_threads: dict[Callable, PooledThread] = dict()

    def __len__(self):
        self._update()
        return len(self.__threads)

    @property
    def threads(self):
        """Number of threads in the pool."""
        self._update()
        return len(self)

    def _update(self):
        """Updates internal variables keeping track of threads in the pool."""
        threads = self.__threads.copy()
        for func, thread in threads.items():
            if not thread.is_alive():
                with self.__lock: self.__dead_threads[func] = self.__threads.pop(func)

        dead_threads = self.__dead_threads.copy()
        for func, thread in dead_threads.items():
            if thread.is_alive():
                with self.__lock: self.__threads[func] = self.__dead_threads.pop(func)

    def thread(self):
        """Function decorator to easily add a function as a thread to the pool."""
        def wrapper(func: Union[Callable, PooledThread], *args, **kwargs):
            return self.add_thread(func, *args, **kwargs)
        return wrapper

    def add_thread(self, func: Union[Callable, PooledThread], *args, **kwargs):
        """Add a function as a thread to the pool."""
        with self.__lock:
            if isinstance(func, PooledThread):
                self.__threads[func._PooledThread__target] = func
            else:
                thread = PooledThread(func, *args, **kwargs)
                self.__threads[func] = thread
                return thread

    def terminate(self, func: Callable = None, block: bool = None, timeout: int = None):
        thread = self.get_thread(func)

        if thread:
            thread.terminate(block=block, timeout=timeout)
            with self.__lock: del self.__threads[func]
            self._update()

    def terminate_all(self, *, block: bool = True, timeout: int = None):
        """Terminates all threads in the pool"""
        for thread in self.__threads.values():
            thread.terminate(block=False)
        self._update()

        if block:
            start = time.time()
            while any(thread.is_alive() for thread in self.__threads.values()):
                if timeout:
                    if time.time() - start > timeout: break
            with self.__lock: self.__threads = dict()

    def pause(self, func: Callable = None, *, resume_in: int = None, block: bool = True, timeout: int = None):
        def resume_dummy():
            time.sleep(resume_in)
            thread.resume()

        thread = self.get_thread(func)

        if thread:
            if resume_in:
                Thread(target=resume_dummy).start()
            thread.pause(block=block, timeout=timeout)
            self._update()

    def pause_all(self, *, resume_in: int = None, block: bool = True, timeout: int = None):
        """Pause all the threads in the pool"""
        def dummy_thread():
            time.sleep(resume_in)
            self.resume_all(block=False)

        for thread in self.__threads.values():
            thread.pause(block=False)
        self._update()

        if resume_in: Thread(target=dummy_thread).start()

        if block:
            start = time.time()
            while any(not thread.is_paused() for thread in self.__threads.values()):
                if timeout:
                    if time.time() - start > timeout: break

    def resume(self, func: Callable = None):
        thread = self.get_thread(func)

        if thread: thread.resume()
        self._update()

    def resume_all(self, *, block: bool = True):
        """Resume all the threads in the pool"""
        for thread in self.__threads.values():
            thread.pause(block=False)
        self._update()

        if block:
            while any(thread.is_paused() for thread in self.__threads.values()): pass

    def get_thread(self, func: Callable = None, checks: list = list(), check: Callable = None) -> Union[PooledThread, None]:
        """Get the `Thread` object from a function."""
        self._update()
        if check: checks.append(check)
        for thread in self.__threads.values():
            if func == thread._PooledThread__target: return thread
            if any(_(func, thread) for _ in checks): return thread



import time
from threading import Condition, Lock, Thread
from typing import Callable


class PooledThread(Thread):
    """A thread that can be paused, resumed and terminated."""
    def __init__(self, target: Callable, *args, **kwargs):
        self.__args = args
        self.__kwargs = kwargs
        self.__target: Callable = target

        self.__lock: Lock = Lock()

        self.__pause: bool = False
        self.__terminate: bool = False
        self.__sleep: bool = False
        self.__execution_event: Condition = Condition(self.__lock)

        super().__init__(target=self.control_wrapper)

    def is_asleep(self):
        """Returns if the thread is sleeping or not."""
        return self.__sleep

    def is_paused(self):
        """Returns if the thread is paused or not."""
        return self.__pause

    def is_alive(self):
        """Returns if the thread is running or not."""
        if self.__sleep or self.__pause: return False
        return super().is_alive()

    def pause(self, *, block: bool = True, timeout: int = None):
        """Pause the execution of thread cycles."""
        with self.__lock: self.__pause = True

        if block: self.wait_cycle_execution(timeout=timeout)

    def resume(self):
        """Resume the execution of thread cycles."""
        with self.__lock: self.__pause = False

    def sleep(self, seconds: int, *, block: bool = False, timeout: int = None):
        """Pause the thread for a given amount of seconds."""
        def dummy_thread():
            time.sleep(seconds)
            with self.__lock: self.__sleep = False

        with self.__lock: self.__sleep = True
        Thread(target=dummy_thread).start()

        if block: self.wait_cycle_execution(timeout=timeout)

    def terminate(self, *, block: bool = True, timeout: int = None):
        """Terminate the thread once the cycle is finished."""
        with self.__lock: self.__terminate = True

        if block: self.wait_cycle_execution(timeout=timeout)

    def wait_cycle_execution(self, *, timeout: int = None):
        """A blocking method that waits till the current thread cycle is over."""
        with self.__lock:
            self.__execution_event.wait(timeout=timeout)

    def control_wrapper(self):
        """A simple function to wrap the original target around to gain control."""
        while not self.__terminate:
            self.__target(*self.__args, **self.__kwargs)
            with self.__lock: self.__execution_event.notify()
            while self.__pause or self.__sleep: pass



import pathlib
import re

from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

packages = [
    "BetterThreads"
]

version = None
with open('BetterThreads/__init__.py') as file:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', file.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError("Version is not set")

setup(
    name="better-threads",
    version=version,
    description="A package that provides extended control over threads",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/ItsYasiru/Better-Threads",
    author="Yasiru",
    author_email="yasiru.dharmathilaka@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9"
    ],
    packages=packages,
    include_package_data=True
)
